{-
	Implementation of Iterated Function System (Barnsley's fern) by Tymon Zgainski
	for 2009 inf1 Haskell Competition
	
	for better performance first compile using ghc fern.hs -O -o fern and then run ./fern
	
	s0951734
	dragomir2@gmail.com
	http://dragomir.org
-}

import System.Random

type Coord = (Float, Float, Int);
type MappedCoord = (Int, Int, Int);

--width = 256
--height = 256

--affine transformations
f1,f2,f3,f4 :: (Float, Float) -> (Float, Float)
f1 (x,y) = (0, 0.16*y)
f2 (x,y) = (0.2*x-0.26*y, 0.23*x+0.22*y+1.6)
f3 (x,y) = (-0.15*x+0.28*y, 0.26*x+0.24*y+0.44)
f4 (x,y) = (0.85*x+0.04*y, -0.04*x+0.85*y+1.6)

initialize :: Int -> Int -> [MappedCoord]
initialize w h = [ (x, y, 0) | x <- [0..w-1], y <- [0..h-1] ]

weights :: [Int]
weights = [ 1 | x <- [1]] ++ [ 2 | x <- [2..8]] ++ [ 3 | x <- [9..15]] ++ [ 4 | x <- [16..100]]

check_weight :: Int -> Int
check_weight x = (if (x == 1) then 1 else if ( (x > 1) && (x < 9) ) then 2 else if ( (x > 8) && (x < 16) ) then 3 else 4)

--generate list of "iterations", each element is number of transformation to use, according to probability of chosing one
toss :: [Int]
--toss = [ check_weight x | (x,y) <- (zip (randomRs (1,100) (mkStdGen 5)) [0..100]) ]
toss = map (\(x,y) -> check_weight x) (zip (randomRs (1,100) (mkStdGen 5)) [0..100000]) --the last number is the number of iterations, change it to change quality and rendering time

-- chaos game, where all the magic happens
chaos :: (Float, Float) -> [Int] -> [Coord]
chaos _ [] = []
chaos (x,y) (z:zs) = [(fst ( v ), snd ( v ), 1)] ++ (chaos v zs)
	where
		v = case z of
			1 -> f1 (x,y)
			2 -> f2 (x,y)
			3 -> f3 (x,y)
			4 -> f4 (x,y)

-- get total hits for each point				
distinct_chaos :: [MappedCoord] -> [MappedCoord]
distinct_chaos [] = []
distinct_chaos ((x,y,f):xs) -- | distinct_chaos ((x,y,f):xs) == distinct_chaos ((x,y,f):xs) = []
							| l /= [] =  [(x,y,1+(sum [ q | (p,r,q) <- l ]))] ++ [ (i,j,k) | (i,j,k) <- distinct_chaos (xs), (i /= x) || (j /= y)] --(filter (\(i,j,k) -> (i /= x) && (j /= y)) distinct_chaos (xs))
							| otherwise = [(x,y,f)] ++ distinct_chaos (xs)
						where l = (filter (\(i,j,k) -> (i == x) && (j == y)) xs)
						
max_x, max_y, min_x, min_y :: [Coord] -> Float
max_x [] = 0;
max_x ((x,y,f):xs) 	| (max_x xs) > x = max_x xs
					| otherwise = x
max_y [] = 0;
max_y ((x,y,f):xs) 	| (max_y xs) > y = max_y xs
					| otherwise = y
min_x [] = 0;
min_x ((x,y,f):xs) 	| (min_x xs) < x = min_x xs
					| otherwise = x
min_y [] = 0;
min_y ((x,y,f):xs) 	| (min_y xs) < y = min_y xs
					| otherwise = y
	
-- not enough time to properly implement scalling, morover there seems to be some kind of bug with other than square images... so it's hardcoded				
max_width, max_height :: [Coord] -> Int
max_width xs = 700--round ((abs (min_x xs) + abs (max_x xs)) / 0.01)
max_height xs = 700--round ((abs (min_y xs) + abs (max_y xs)) / 0.01)

max_width_u, max_height_u :: [Coord] -> Float
max_width_u xs = ((abs (min_x xs) + abs (max_x xs)))
max_height_u xs = ((abs (min_y xs) + abs (max_y xs)))
			
map_chaos :: [Coord] -> [MappedCoord]		
map_chaos xs =  [ ( round(x * 64) + 350, round(y * 64) + 44, f) | (x,y,f) <- xs ]
--[ ( round((x * fromIntegral(max_width xs)) / (max_width_u xs)), round((y * fromIntegral(max_height xs)) / (max_height_u xs)), f) | (x,y,f) <- xs ]
--[ ( round( ((x + 2.1818) / 4.8374) * 483.74/2), round( ((y + 0.0) / 9.95851) * 995.851/2), f) | (x,y,f) <- xs ]
--[ ( round((x + abs(min_x xs)) / 0.01), round((y + abs(min_y xs)) / 0.01), f) | (x,y,f) <- xs ]

--sort each point and calculate its colour
sort_chaos :: [MappedCoord] -> [MappedCoord] -> [MappedCoord]
sort_chaos _ [] = []
sort_chaos (c:cs) ((x,y,f):xs) 	| length z > 1 = error "something went terribly wrong"
								| z == [] = [(x,y,0)] ++ (sort_chaos (c:cs) xs)
								| otherwise = (map (\(k,l,m) -> (x,y, (m * 20) )) z) ++ (sort_chaos (c:cs) xs)
							where z = [ (k,l,m) | (k,l,m) <- (c:cs), (k == x) && (l == y) ]
	


--set = [ (x, y, r) | 

serialize :: MappedCoord -> String
serialize (x,y,f) = (show 0) ++ " " ++ (show (f)) ++ " " ++ (show 0)

ppm :: [MappedCoord] -> Int -> Int -> String
ppm set w h = "P3\n" ++ (show (w)) ++ "\n" ++ (show (h)) ++ "\n" ++ "255" ++ "\n" ++ (foldr (++) [] ( [ (serialize x) ++ "\n" | x <- set ] ) )

main = writeFile "out.ppm" $ ppm (sort_chaos (distinct_chaos (map_chaos set)) (initialize (max_height set) (max_height set))) (max_width set) (max_height set)
	where set = (chaos (0,0) toss)


